/*
 *  Copyright (C) 2002-2010  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: drive_local.cpp,v 1.82 2009-07-18 18:42:55 c2woody Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <errno.h>
#include <cassert>
#include "dosbox.h"
#include "dos_inc.h"
#include "drives.h"
#include "support.h"
#include "cross.h"
#include "inout.h"

namespace {
class localFile : public DOS_File {
public:
	localFile(const char* name, FILE * handle);
	bool Read(Bit8u * data,Bit16u * size);
	bool Write(Bit8u * data,Bit16u * size);
	bool Seek(Bit32u * pos,Bit32u type);
	bool Commit() override;
	bool Close();
	bool LockFile(Bit8u mode, Bit32u pos, Bit32u size);
	Bit16u GetInformation(void);
	bool UpdateDateTimeFromHost(void);   
	void FlagReadOnlyMedium(void);
	void Flush(void);
private:
	FILE * fhandle;
	bool read_only_medium;
	enum { NONE,READ,WRITE } last_action;
};


std::string FullPath(std::string const& basedir, char const* const path) {
    assert(!basedir.empty());
    assert(path);
    if (basedir.empty()) return path;

    auto full = [&] {
        // Empty path, trim off the last \ or / since a few funcs are bitchy about
        // having a dangling separator.
        if (!(path && *path)) {
            assert((basedir.back() == '\\') ||
                   (basedir.back() == '/' ));
            return basedir.substr(0, basedir.length() - 1);
        }

        return basedir + path;
    }();

    CROSS_FILENAME(full);
    return full;
}

}


bool localDrive::FileCreate(DOS_File * * file,char * name,Bit16u /*attributes*/) {
//TODO Maybe care for attributes but not likely
	auto newname = FullPath(basedir, name);
	/* Test if file exists (so we need to truncate it). don't add to dirCache then */
	bool existing_file=false;
	
	FILE * test = fopen(newname.c_str(), "rb+");
	if(test) {
		fclose(test);
		existing_file=true;

	}
	
	FILE * hand = fopen(newname.c_str(), "wb+");
	if (!hand){
		LOG_MSG("Warning: file creation failed: %s",newname);
		return false;
	}

	/* Make the 16 bit device information */
	*file=new localFile(name,hand);
	(*file)->flags=OPEN_READWRITE;

	return true;
}

bool localDrive::FileOpen(DOS_File * * file,char * name,Bit32u flags) {
	const char* type;
	switch (flags&0xf) {
	case OPEN_READ:type="rb"; break;
	case OPEN_WRITE:type="rb+"; break;
	case OPEN_READWRITE:type="rb+"; break;
	default:
		DOS_SetError(DOSERR_ACCESS_CODE_INVALID);
		return false;
	}
	auto newname = FullPath(basedir, name);

	//Flush the buffer of handles for the same file. (Betrayal in Antara)
	Bit8u i,drive=DOS_DRIVES;
	localFile *lfp;
	for (i=0;i<DOS_DRIVES;i++) {
		if (Drives[i]==this) {
			drive=i;
			break;
		}
	}
	for (i=0;i<DOS_FILES;i++) {
		if (Files[i] && Files[i]->IsOpen() && Files[i]->GetDrive()==drive && Files[i]->IsName(name)) {
			lfp=dynamic_cast<localFile*>(Files[i]);
			if (lfp) lfp->Flush();
		}
	}

	FILE * hand=fopen(newname.c_str(),type);
//	Bit32u err=errno;
	if (!hand) { 
		if((flags&0xf) != OPEN_READ) {
			FILE * hmm=fopen(newname.c_str(),"rb");
			if (hmm) {
				fclose(hmm);
				LOG_MSG("Warning: file %s exists and failed to open in write mode.\nPlease Remove write-protection",newname);
			}
		}
		return false;
	}

	*file=new localFile(name,hand);
	(*file)->flags=flags;  //for the inheritance flag and maybe check for others.
//	(*file)->SetFileName(newname);
	return true;
}

FILE * localDrive::GetSystemFilePtr(char const * const name, char const * const type) {
	auto newname = FullPath(basedir, name);

	return fopen(newname.c_str(),type);
}

std::string localDrive::GetSystemFilename(char const * const dosName) {
	std::string  newname = basedir + dosName;
	CROSS_FILENAME(sysName);
	return newname;
}

bool localDrive::FileUnlink(char * name) {
	auto newname = FullPath(basedir, name);
	if (!unlink(newname.c_str())) return true;

	//Unlink failed for some reason try finding it.
	if (!CrossFileOrDirExists(newname.c_str())) return false; // File not found.

	FILE* file_writable = fopen(newname.c_str(), "rb+");
	if(!file_writable) return false; //No acces ? ERROR MESSAGE NOT SET. FIXME ?
	fclose(file_writable);

	//File exists and can technically be deleted, nevertheless it failed.
	//This means that the file is probably open by some process.
	//See if We have it open.
	bool found_file = false;
	for(Bitu i = 0;i < DOS_FILES;i++){
		if(Files[i] && Files[i]->IsName(name)) {
			Bitu max = DOS_FILES;
			while(Files[i]->IsOpen() && max--) {
				Files[i]->Close();
				if (Files[i]->RemoveRef()<=0) break;
			}
			found_file=true;
		}
	}
	return found_file && (!unlink(newname.c_str()));
}

namespace {
localDrive::SrchInfo& SearchInfoSlotAlloc(localDrive::SrchInfos& a) {
	assert(a.possibleNextFree < a.infos.max_size());

	auto& slot = [&]() -> localDrive::SrchInfo& {
		if (a.infos[a.possibleNextFree].hFind == INVALID_HANDLE_VALUE)
			return a.infos[a.possibleNextFree];

		for (auto& i : a.infos)
			if (i.hFind == INVALID_HANDLE_VALUE) return i;

		LOG(LOG_MISC, LOG_ERROR)("localDrive-Find: All slots full. Picking next possible.");
		return a.infos[a.possibleNextFree]; // Give up. Roll next-override forward.
	}();

	auto const idx      = &slot - &a.infos.front();
	a.possibleNextFree  = (idx + 1) % a.infos.max_size();
	return slot;
}
}

bool localDrive::FindFirst(char * _dir,DOS_DTA & dta,bool fcb_findfirst) {
	if (!TestDir(_dir)) {
		DOS_SetError(DOSERR_PATH_NOT_FOUND);
		return false;
	}

	std::string tempDir = basedir + _dir;
	CROSS_FILENAME(tempDir);

	if (allocation.mediaid==0xF0 ) {
		EmptyCache(); //rescan floppie-content on each findfirst
	}

	auto& info = SearchInfoSlotAlloc(srchInfos);
	if (info.hFind != INVALID_HANDLE_VALUE) {
		LOG(LOG_MISC, LOG_ERROR)("localDrive: FindFirst/Next: All slots full. Picking first.");
		::FindClose(info.hFind);
		info.hFind = INVALID_HANDLE_VALUE;
	}

	if (tempDir[tempDir.length() - 1] != CROSS_FILESPLIT)
		tempDir += CROSS_FILESPLIT;
	info.dirAbs = std::move(tempDir);
	dta.SetDirID(&info - &srchInfos.infos.front());

	Bit8u sAttr; char srch_pattern[DOS_NAMELENGTH_ASCII];
	dta.GetSearchParams(sAttr, srch_pattern);

	if (this->isRemote() && this->isRemovable()) {
		// cdroms behave a bit different than regular drives
		if (sAttr == DOS_ATTR_VOLUME) {
			dta.SetResult(GetLabel(),0,0,0,DOS_ATTR_VOLUME);
			return true;
		}
	} else {
		if (sAttr == DOS_ATTR_VOLUME) {
			if ( strcmp(GetLabel(), "") == 0 ) {
//				LOG(LOG_DOSMISC,LOG_ERROR)("DRIVELABEL REQUESTED: none present, returned  NOLABEL");
//				dta.SetResult("NO_LABEL",0,0,0,DOS_ATTR_VOLUME);
//				return true;
				DOS_SetError(DOSERR_NO_MORE_FILES);
				return false;
			}
			dta.SetResult(GetLabel(),0,0,0,DOS_ATTR_VOLUME);
			return true;
		} else if ((sAttr & DOS_ATTR_VOLUME)  && (*_dir == 0) && !fcb_findfirst) { 
		//should check for a valid leading directory instead of 0
		//exists==true if the volume label matches the searchmask and the path is valid
			if (WildFileCmp(GetLabel(), srch_pattern)) {
				dta.SetResult(GetLabel(),0,0,0,DOS_ATTR_VOLUME);
				return true;
			}
		}
	}
	return FindNext(dta);
}


bool localDrive::FindNext(DOS_DTA& dta) {
	Bit8u srch_attr; char srch_pattern[DOS_NAMELENGTH_ASCII];
	dta.GetSearchParams(srch_attr, srch_pattern);

	auto const FileNameDOS = [](char const* const s) {
		if (!s                      ) return true;
		if (strcmp(s, "."   ) == 0  ) return true;
		if (strcmp(s, ".."  ) == 0  ) return true;

		auto const len          = strlen(s);
		auto const dot          = strchr(s, '.');
		auto const bodyLen      = dot ? (dot - s) : len;
		auto const dotExtLen    = len - bodyLen;
		if (bodyLen   > 8               ) return false;
		if (dotExtLen > 4               ) return false;
		if (dot && strchr(dot + 1, '.') ) return false; // Two dots not permitted unless ".."

		auto const kBadChars    = "\"*+,/:;<=>?\\[]|";
		if (strpbrk(s, kBadChars)       ) return false;

		return true;
	};

	auto const Skip = [&](std::string const& pathAbs, WIN32_FIND_DATA const& wfd
													, struct stat&           stat_block) {
		auto const nameSep      = strrchr(pathAbs.c_str(), CROSS_FILESPLIT); assert(nameSep);
		if (!nameSep) return true;

		// This can still fail if our attempts to generate a short file name failed for whatever
		// reason. This can't be fixed reliably if we don't have a persistent store for the short
		// file names.
		auto const nameShort = nameSep + 1;
		if (!FileNameDOS(nameShort              )) return true;
		if (!WildFileCmp(nameShort, srch_pattern)) return true;

		if (stat(pathAbs.c_str(), &stat_block) != 0) return true; // No symlinks and such

		Bit8u const find_attr = (stat_block.st_mode & S_IFDIR) ? DOS_ATTR_DIRECTORY
			: DOS_ATTR_ARCHIVE;
		if (~srch_attr & find_attr & (DOS_ATTR_DIRECTORY | DOS_ATTR_HIDDEN | DOS_ATTR_SYSTEM))
			return true;

		return false;
	};

	auto const Bind = [&](WIN32_FIND_DATA const& wfd, struct stat const& stat_block) {
		auto const nameShort    = *wfd.cAlternateFileName   ? wfd.cAlternateFileName
															: wfd.cFileName;
		auto const find_attr    = (stat_block.st_mode & S_IFDIR) ? DOS_ATTR_DIRECTORY
																 : DOS_ATTR_ARCHIVE;

		Bit16u find_date = 6, find_time = 4;
		if (auto const time = localtime(&stat_block.st_mtime)) {
			find_date = DOS_PackDate((Bit16u)(time->tm_year + 1900), (Bit16u)(time->tm_mon + 1)
									,(Bit16u) time->tm_mday );
			find_time = DOS_PackTime((Bit16u) time->tm_hour        , (Bit16u) time->tm_min
									,(Bit16u) time->tm_sec  );
		}

		char find_name[DOS_NAMELENGTH_ASCII];
		strcpy_s(find_name, nameShort); upcase(find_name);

		dta.SetResult(find_name, (Bit32u)stat_block.st_size, find_date, find_time, find_attr);
	};

	auto const Search = [&](SrchInfo& i, WIN32_FIND_DATA& wfd, struct stat& stat_block) {
		std::string pathAbs;
		do {
			if (i.hFind == INVALID_HANDLE_VALUE)
				i.hFind = ::FindFirstFileEx((i.dirAbs + '*').c_str(), FindExInfoStandard, &wfd,
								FindExSearchNameMatch, nullptr, FIND_FIRST_EX_LARGE_FETCH);
			else if (!::FindNextFile(i.hFind, &wfd))
				i.hFind = INVALID_HANDLE_VALUE;

			if (i.hFind == INVALID_HANDLE_VALUE) return false;

			auto const s = *wfd.cAlternateFileName ? wfd.cAlternateFileName : wfd.cFileName;
			pathAbs = i.dirAbs + s;
			if (!FileNameDOS(s)) { // Can happen if a remote store, for example
				auto const bufSz = GetShortPathName(pathAbs.c_str(), nullptr, 0);
				if (bufSz == 0) {
					pathAbs.clear(); // Failed. Clear abs path. Filter will ignore.
					continue;
				}

				std::vector<char> tmp(bufSz);
				if (GetShortPathName(pathAbs.c_str(), tmp.data(), tmp.size()) != (bufSz - 1)) {
					pathAbs.clear(); // Failed. Clear abs path. Filter will ignore.
					continue;
				}

				pathAbs = tmp.data();
			}
		} while (Skip(pathAbs, wfd, stat_block));

		
		return true;
	};

	WIN32_FIND_DATA wfd; struct stat stat_block;
	if (!Search(srchInfos.infos[dta.GetDirID()], wfd, stat_block)) {
		DOS_SetError(DOSERR_NO_MORE_FILES);
		return false;
	}

	Bind(wfd, stat_block);
	return true;
}

bool localDrive::GetFileAttr(char * name,Bit16u * attr) {
	auto const newname  = FullPath(basedir, name);
    auto const fileAttr = ::GetFileAttributes(newname.c_str());
	if (fileAttr != INVALID_FILE_ATTRIBUTES) {
		*attr=DOS_ATTR_ARCHIVE;
		if(fileAttr & FILE_ATTRIBUTE_DIRECTORY) *attr|=DOS_ATTR_DIRECTORY;
		return true;
	}
	*attr=0;
	return false; 
}

bool localDrive::MakeDir(char * dir) {
	auto newdir = FullPath(basedir, dir);
#if defined (WIN32)						/* MS Visual C++ */
	int temp=mkdir(newdir.c_str());
#else
	int temp=mkdir(newdir.c_str(),0700);
#endif
	return (temp==0);// || ((temp!=0) && (errno==EEXIST));
}

bool localDrive::RemoveDir(char * dir) {
	auto newdir = FullPath(basedir, dir);
	return rmdir(newdir.c_str()) == 0;
}

bool localDrive::TestDir(char * dir) {
	auto const newdir   = FullPath(basedir, dir);
	auto const fileAttr = ::GetFileAttributes(newdir.c_str());
	if (fileAttr == INVALID_FILE_ATTRIBUTES     ) return false;
	if (!(fileAttr & FILE_ATTRIBUTE_DIRECTORY)  ) return false;

	return access(newdir.c_str(),F_OK) == 0;
}

bool localDrive::Rename(char * oldname,char * newname) {
	std::string newold = basedir + oldname;
	std::string newnew = basedir + newname;
	CROSS_FILENAME(newold);
	CROSS_FILENAME(newnew);
	return rename(newold.c_str(), newnew.c_str()) == 0;
}

bool localDrive::AllocationInfo(Bit16u * _bytes_sector,Bit8u * _sectors_cluster,Bit16u * _total_clusters,Bit16u * _free_clusters) {
	/* Always report 100 mb free should be enough */
	/* Total size is always 1 gb */
	*_bytes_sector=allocation.bytes_sector;
	*_sectors_cluster=allocation.sectors_cluster;
	*_total_clusters=allocation.total_clusters;
	*_free_clusters=allocation.free_clusters;
	return true;
}

bool localDrive::FileExists(const char* name) {
	auto newname = FullPath(basedir, name);
	FILE* Temp=fopen(newname.c_str(),"rb");
	if(Temp==NULL) return false;
	fclose(Temp);
	return true;
}

bool localDrive::FileStat(const char* name, FileStat_Block * const stat_block) {
	auto newname = FullPath(basedir, name);
    struct stat temp_stat { 0 };
    /// \fixme Will this work for for remote mounts?
	if(stat(newname.c_str(),&temp_stat)!=0) return false;
	/* Convert the stat to a FileStat */
	struct tm *time;
	if((time=localtime(&temp_stat.st_mtime))!=0) {
		stat_block->time=DOS_PackTime((Bit16u)time->tm_hour,(Bit16u)time->tm_min,(Bit16u)time->tm_sec);
		stat_block->date=DOS_PackDate((Bit16u)(time->tm_year+1900),(Bit16u)(time->tm_mon+1),(Bit16u)time->tm_mday);
	} else {

	}
	stat_block->size=(Bit32u)temp_stat.st_size;
	return true;
}


Bit8u localDrive::GetMediaByte(void) {
	return allocation.mediaid;
}

bool localDrive::isRemote(void) {
	return false;
}

bool localDrive::isRemovable(void) {
	return false;
}

Bits localDrive::UnMount(void) { 
	delete this;
	return 0; 
}

localDrive::localDrive(const char * startdir,Bit16u _bytes_sector,Bit8u _sectors_cluster,Bit16u _total_clusters,Bit16u _free_clusters,Bit8u _mediaid) {
	basedir = startdir;
	info = std::string("local directory ") + startdir;
	allocation.bytes_sector=_bytes_sector;
	allocation.sectors_cluster=_sectors_cluster;
	allocation.total_clusters=_total_clusters;
	allocation.free_clusters=_free_clusters;
	allocation.mediaid=_mediaid;
}


//TODO Maybe use fflush, but that seemed to fuck up in visual c
bool localFile::Read(Bit8u * data,Bit16u * size) {
	if ((this->flags & 0xf) == OPEN_WRITE) {	// check if file opened in write-only mode
		DOS_SetError(DOSERR_ACCESS_DENIED);
		return false;
	}
	if (last_action==WRITE) fseek(fhandle,ftell(fhandle),SEEK_SET);
	last_action=READ;
	*size=(Bit16u)fread(data,1,*size,fhandle);
	/* Fake harddrive motion. Inspector Gadget with soundblaster compatible */
	/* Same for Igor */
	/* hardrive motion => unmask irq 2. Only do it when it's masked as unmasking is realitively heavy to emulate */
	Bit8u mask = IO_Read(0x21);
	if(mask & 0x4 ) IO_Write(0x21,mask&0xfb);
	return true;
}

bool localFile::Write(Bit8u * data,Bit16u * size) {
	if ((this->flags & 0xf) == OPEN_READ) {	// check if file opened in read-only mode
		DOS_SetError(DOSERR_ACCESS_DENIED);
		return false;
	}
	if (last_action==READ) fseek(fhandle,ftell(fhandle),SEEK_SET);
	last_action=WRITE;
	if(*size==0){  
		return (!ftruncate(fileno(fhandle),ftell(fhandle)));
	}
	else 
	{
		*size=(Bit16u)fwrite(data,1,*size,fhandle);
		return true;
	}
}

/* ert, 20100711: Locking extensions */
#ifdef WIN32
#include <sys/locking.h>
bool localFile::LockFile(Bit8u const mode, Bit32u const pos, Bit32u const size) {
	auto const hFile = (HANDLE)_get_osfhandle(_fileno(fhandle));

	switch (mode) {
	case 0  : { if (::LockFile  (hFile, pos, 0, size, 0)) return true; } break;
	case 1  : { if (::UnlockFile(hFile, pos, 0, size, 0)) return true; } break;
	default : {
		DOS_SetError(DOSERR_FUNCTION_NUMBER_INVALID);
	} return false;
	}

	switch (GetLastError()) {
	case ERROR_ACCESS_DENIED        : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_LOCK_FAILED          : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_NETWORK_ACCESS_DENIED: // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_DRIVE_LOCKED         : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_SEEK_ON_DEVICE       : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_NOT_LOCKED           : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_LOCK_VIOLATION       : { DOS_SetError(DOSERR_LOCK_VIOLATION          ); } break;
	case ERROR_INVALID_HANDLE       : { DOS_SetError(DOSERR_INVALID_HANDLE          ); } break;
	case ERROR_INVALID_FUNCTION     : // FALL TO: Default
	default                         : { DOS_SetError(DOSERR_FUNCTION_NUMBER_INVALID ); } break;
	}

	return false;
}

bool localFile::Commit() {
	auto const hFile = (HANDLE)_get_osfhandle(_fileno(fhandle));
	if (::FlushFileBuffers(hFile)) return true;

	switch (GetLastError()) {
	case ERROR_ACCESS_DENIED        : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_LOCK_FAILED          : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_NETWORK_ACCESS_DENIED: // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_DRIVE_LOCKED         : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_SEEK_ON_DEVICE       : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_NOT_LOCKED           : // FALL TO: ERROR_LOCK_VIOLATION
	case ERROR_LOCK_VIOLATION       : { DOS_SetError(DOSERR_LOCK_VIOLATION          ); } break;
	case ERROR_INVALID_HANDLE       : { DOS_SetError(DOSERR_INVALID_HANDLE          ); } break;
	case ERROR_INVALID_FUNCTION     : // FALL TO: Default
	default                         : { DOS_SetError(DOSERR_FUNCTION_NUMBER_INVALID ); } break;
	}

	return false;
}
#endif

bool localFile::Seek(Bit32u * pos,Bit32u type) {
	int seektype;
	switch (type) {
	case DOS_SEEK_SET:seektype=SEEK_SET;break;
	case DOS_SEEK_CUR:seektype=SEEK_CUR;break;
	case DOS_SEEK_END:seektype=SEEK_END;break;
	default:
	//TODO Give some doserrorcode;
		return false;//ERROR
	}
	int ret=fseek(fhandle,*reinterpret_cast<Bit32s*>(pos),seektype);
	if (ret!=0) {
		// Out of file range, pretend everythings ok 
		// and move file pointer top end of file... ?! (Black Thorne)
		fseek(fhandle,0,SEEK_END);
	};
#if 0
	fpos_t temppos;
	fgetpos(fhandle,&temppos);
	Bit32u * fake_pos=(Bit32u*)&temppos;
	*pos=*fake_pos;
#endif
	*pos=(Bit32u)ftell(fhandle);
	last_action=NONE;
	return true;
}

bool localFile::Close() {
	// only close if one reference left
	if (refCtr==1) {
		if(fhandle) fclose(fhandle);
		fhandle = 0;
		open = false;
	};
	return true;
}

Bit16u localFile::GetInformation(void) {
	return read_only_medium?0x40:0;
}
	

localFile::localFile(const char* _name, FILE * handle) {
	fhandle=handle;
	open=true;
	UpdateDateTimeFromHost();

	attr=DOS_ATTR_ARCHIVE;
	last_action=NONE;
	read_only_medium=false;

	name=0;
	SetName(_name);
}

void localFile::FlagReadOnlyMedium(void) {
	read_only_medium = true;
}

bool localFile::UpdateDateTimeFromHost(void) {
	if(!open) return false;
	struct stat temp_stat;
	fstat(fileno(fhandle),&temp_stat);
	struct tm * ltime;
	if((ltime=localtime(&temp_stat.st_mtime))!=0) {
		time=DOS_PackTime((Bit16u)ltime->tm_hour,(Bit16u)ltime->tm_min,(Bit16u)ltime->tm_sec);
		date=DOS_PackDate((Bit16u)(ltime->tm_year+1900),(Bit16u)(ltime->tm_mon+1),(Bit16u)ltime->tm_mday);
	} else {
		time=1;date=1;
	}
	return true;
}

void localFile::Flush(void) {
	if (last_action==WRITE) {
		fseek(fhandle,ftell(fhandle),SEEK_SET);
		last_action=NONE;
	}
}


// ********************************************
// CDROM DRIVE
// ********************************************

int  MSCDEX_RemoveDrive(char driveLetter);
int  MSCDEX_AddDrive(char driveLetter, const char* physicalPath, Bit8u& subUnit);
bool MSCDEX_HasMediaChanged(Bit8u subUnit);
bool MSCDEX_GetVolumeName(Bit8u subUnit, char* name);

namespace {
bool LabelFetch(Bit8u const subUnit, std::string& output) {
	char name[32];
	if (!MSCDEX_GetVolumeName(subUnit, name)) return false;

	char nameTmp[512]; Set_Label(name, nameTmp, true);
	output = nameTmp;
	return true;
}
}

cdromDrive::cdromDrive(const char driveLetter, const char * startdir,Bit16u _bytes_sector,Bit8u _sectors_cluster,Bit16u _total_clusters,Bit16u _free_clusters,Bit8u _mediaid, int& error)
		   :localDrive(startdir,_bytes_sector,_sectors_cluster,_total_clusters,_free_clusters,_mediaid) {
	// Init mscdex
	error = MSCDEX_AddDrive(driveLetter,startdir,subUnit);
	info = std::string("CDRom ") + startdir;
	this->driveLetter = driveLetter;
	// Get Volume Label
	LabelFetch(subUnit, label);
}

bool cdromDrive::FileOpen(DOS_File * * file,char * name,Bit32u flags) {
	if ((flags&0xf)==OPEN_READWRITE) {
		flags &= ~OPEN_READWRITE;
	} else if ((flags&0xf)==OPEN_WRITE) {
		DOS_SetError(DOSERR_ACCESS_DENIED);
		return false;
	}
	bool retcode = localDrive::FileOpen(file,name,flags);
	if(retcode) (dynamic_cast<localFile*>(*file))->FlagReadOnlyMedium();
	return retcode;
}

bool cdromDrive::FileCreate(DOS_File * * /*file*/,char * /*name*/,Bit16u /*attributes*/) {
	DOS_SetError(DOSERR_ACCESS_DENIED);
	return false;
}

bool cdromDrive::FileUnlink(char * /*name*/) {
	DOS_SetError(DOSERR_ACCESS_DENIED);
	return false;
}

bool cdromDrive::RemoveDir(char * /*dir*/) {
	DOS_SetError(DOSERR_ACCESS_DENIED);
	return false;
}

bool cdromDrive::MakeDir(char * /*dir*/) {
	DOS_SetError(DOSERR_ACCESS_DENIED);
	return false;
}

bool cdromDrive::Rename(char * /*oldname*/,char * /*newname*/) {
	DOS_SetError(DOSERR_ACCESS_DENIED);
	return false;
}

bool cdromDrive::GetFileAttr(char * name,Bit16u * attr) {
	bool result = localDrive::GetFileAttr(name,attr);
	if (result) *attr |= DOS_ATTR_READ_ONLY;
	return result;
}

bool cdromDrive::FindFirst(char * _dir,DOS_DTA & dta,bool /*fcb_findfirst*/) {
	// If media has changed, reInit drivecache.
	if (MSCDEX_HasMediaChanged(subUnit)) {
		// Get Volume Label
		LabelFetch(subUnit, label);
	}
	return localDrive::FindFirst(_dir,dta);
}

void cdromDrive::SetDir(const char* path) {
	// If media has changed, reInit drivecache.
	if (MSCDEX_HasMediaChanged(subUnit)) {
		// Get Volume Label
		LabelFetch(subUnit, label);
	}
	localDrive::SetDir(path);
}

bool cdromDrive::isRemote(void) {
	return true;
}

bool cdromDrive::isRemovable(void) {
	return true;
}

Bits cdromDrive::UnMount(void) {
	if(MSCDEX_RemoveDrive(driveLetter)) {
		delete this;
		return 0;
	}
	return 2;
}
