#include <cassert>
#include <string>
#include <queue>

#include "dosbox.h"
#include "logging.h"
#include "sysmenu.h"
#include "keyboard.h"
#include "SDL.h"
#include "SDL_syswm.h"
#include "parport.h"
#include "hardware.h"
#include "TCPSocket.h"

#define WM_PRINTERREDIRECT_PORTNOTIFY (WM_USER + 1)

#if defined(WIN32)
    // Ripped from SDL's SDL_dx5events.c, since there's no API to access it...
    #include <dinput.h>
    #ifndef DIK_PAUSE
        #define DIK_PAUSE	0xC5
    #endif
    #ifndef DIK_OEM_102
        #define DIK_OEM_102	0x56	/* < > | on UK/Germany keyboards */
    #endif
#else
    #error Impl pasting for non-win32
#endif

#ifndef LENGTH_OF
    #define LENGTH_OF(a) (sizeof(a) / sizeof(*a))
#endif


void MAPPER_CheckEvent(SDL_Event* event);

namespace {
    char const* const szAboutFile = "about.txt";
    size_t      const kPasteMinBufExtra = 4; //< Just in case, to keep us from entering an unexpected KB state
    std::queue<char>  pasteBuffer;
    bool              setupDone           = false;
    bool              printerRedirect     = false;
    uint32_t          printerRedirectPort = 0;
    uint32_t          printScreenPort     = 0;
    
    decltype(SDL_SysWMinfo::window) WindowGet() {
        SDL_SysWMinfo wmiInfo; SDL_VERSION(&wmiInfo.version);
        return (SDL_GetWMInfo(&wmiInfo) == 1) ? wmiInfo.window : NULL;
    }
    
#if defined(WIN32)
    WNDPROC fnSDLDefaultWndProc       = nullptr;
    SDLKey  aryScanCodeToSDLKey[0xFF] = {SDLK_UNKNOWN};
    
    void PasteInitMapSCToSDLKey()
    {
        static bool bScanCodeMapInited = false;
        if (bScanCodeMapInited) return;
        bScanCodeMapInited = true;
        
        /* Defined DIK_* constants */
        aryScanCodeToSDLKey[DIK_ESCAPE] = SDLK_ESCAPE;
        aryScanCodeToSDLKey[DIK_1] = SDLK_1;
        aryScanCodeToSDLKey[DIK_2] = SDLK_2;
        aryScanCodeToSDLKey[DIK_3] = SDLK_3;
        aryScanCodeToSDLKey[DIK_4] = SDLK_4;
        aryScanCodeToSDLKey[DIK_5] = SDLK_5;
        aryScanCodeToSDLKey[DIK_6] = SDLK_6;
        aryScanCodeToSDLKey[DIK_7] = SDLK_7;
        aryScanCodeToSDLKey[DIK_8] = SDLK_8;
        aryScanCodeToSDLKey[DIK_9] = SDLK_9;
        aryScanCodeToSDLKey[DIK_0] = SDLK_0;
        aryScanCodeToSDLKey[DIK_MINUS] = SDLK_MINUS;
        aryScanCodeToSDLKey[DIK_EQUALS] = SDLK_EQUALS;
        aryScanCodeToSDLKey[DIK_BACK] = SDLK_BACKSPACE;
        aryScanCodeToSDLKey[DIK_TAB] = SDLK_TAB;
        aryScanCodeToSDLKey[DIK_Q] = SDLK_q;
        aryScanCodeToSDLKey[DIK_W] = SDLK_w;
        aryScanCodeToSDLKey[DIK_E] = SDLK_e;
        aryScanCodeToSDLKey[DIK_R] = SDLK_r;
        aryScanCodeToSDLKey[DIK_T] = SDLK_t;
        aryScanCodeToSDLKey[DIK_Y] = SDLK_y;
        aryScanCodeToSDLKey[DIK_U] = SDLK_u;
        aryScanCodeToSDLKey[DIK_I] = SDLK_i;
        aryScanCodeToSDLKey[DIK_O] = SDLK_o;
        aryScanCodeToSDLKey[DIK_P] = SDLK_p;
        aryScanCodeToSDLKey[DIK_LBRACKET] = SDLK_LEFTBRACKET;
        aryScanCodeToSDLKey[DIK_RBRACKET] = SDLK_RIGHTBRACKET;
        aryScanCodeToSDLKey[DIK_RETURN] = SDLK_RETURN;
        aryScanCodeToSDLKey[DIK_LCONTROL] = SDLK_LCTRL;
        aryScanCodeToSDLKey[DIK_A] = SDLK_a;
        aryScanCodeToSDLKey[DIK_S] = SDLK_s;
        aryScanCodeToSDLKey[DIK_D] = SDLK_d;
        aryScanCodeToSDLKey[DIK_F] = SDLK_f;
        aryScanCodeToSDLKey[DIK_G] = SDLK_g;
        aryScanCodeToSDLKey[DIK_H] = SDLK_h;
        aryScanCodeToSDLKey[DIK_J] = SDLK_j;
        aryScanCodeToSDLKey[DIK_K] = SDLK_k;
        aryScanCodeToSDLKey[DIK_L] = SDLK_l;
        aryScanCodeToSDLKey[DIK_SEMICOLON] = SDLK_SEMICOLON;
        aryScanCodeToSDLKey[DIK_APOSTROPHE] = SDLK_QUOTE;
        aryScanCodeToSDLKey[DIK_GRAVE] = SDLK_BACKQUOTE;
        aryScanCodeToSDLKey[DIK_LSHIFT] = SDLK_LSHIFT;
        aryScanCodeToSDLKey[DIK_BACKSLASH] = SDLK_BACKSLASH;
        aryScanCodeToSDLKey[DIK_OEM_102] = SDLK_LESS;
        aryScanCodeToSDLKey[DIK_Z] = SDLK_z;
        aryScanCodeToSDLKey[DIK_X] = SDLK_x;
        aryScanCodeToSDLKey[DIK_C] = SDLK_c;
        aryScanCodeToSDLKey[DIK_V] = SDLK_v;
        aryScanCodeToSDLKey[DIK_B] = SDLK_b;
        aryScanCodeToSDLKey[DIK_N] = SDLK_n;
        aryScanCodeToSDLKey[DIK_M] = SDLK_m;
        aryScanCodeToSDLKey[DIK_COMMA] = SDLK_COMMA;
        aryScanCodeToSDLKey[DIK_PERIOD] = SDLK_PERIOD;
        aryScanCodeToSDLKey[DIK_SLASH] = SDLK_SLASH;
        aryScanCodeToSDLKey[DIK_RSHIFT] = SDLK_RSHIFT;
        aryScanCodeToSDLKey[DIK_MULTIPLY] = SDLK_KP_MULTIPLY;
        aryScanCodeToSDLKey[DIK_LMENU] = SDLK_LALT;
        aryScanCodeToSDLKey[DIK_SPACE] = SDLK_SPACE;
        aryScanCodeToSDLKey[DIK_CAPITAL] = SDLK_CAPSLOCK;
        aryScanCodeToSDLKey[DIK_F1] = SDLK_F1;
        aryScanCodeToSDLKey[DIK_F2] = SDLK_F2;
        aryScanCodeToSDLKey[DIK_F3] = SDLK_F3;
        aryScanCodeToSDLKey[DIK_F4] = SDLK_F4;
        aryScanCodeToSDLKey[DIK_F5] = SDLK_F5;
        aryScanCodeToSDLKey[DIK_F6] = SDLK_F6;
        aryScanCodeToSDLKey[DIK_F7] = SDLK_F7;
        aryScanCodeToSDLKey[DIK_F8] = SDLK_F8;
        aryScanCodeToSDLKey[DIK_F9] = SDLK_F9;
        aryScanCodeToSDLKey[DIK_F10] = SDLK_F10;
        aryScanCodeToSDLKey[DIK_NUMLOCK] = SDLK_NUMLOCK;
        aryScanCodeToSDLKey[DIK_SCROLL] = SDLK_SCROLLOCK;
        aryScanCodeToSDLKey[DIK_NUMPAD7] = SDLK_KP7;
        aryScanCodeToSDLKey[DIK_NUMPAD8] = SDLK_KP8;
        aryScanCodeToSDLKey[DIK_NUMPAD9] = SDLK_KP9;
        aryScanCodeToSDLKey[DIK_SUBTRACT] = SDLK_KP_MINUS;
        aryScanCodeToSDLKey[DIK_NUMPAD4] = SDLK_KP4;
        aryScanCodeToSDLKey[DIK_NUMPAD5] = SDLK_KP5;
        aryScanCodeToSDLKey[DIK_NUMPAD6] = SDLK_KP6;
        aryScanCodeToSDLKey[DIK_ADD] = SDLK_KP_PLUS;
        aryScanCodeToSDLKey[DIK_NUMPAD1] = SDLK_KP1;
        aryScanCodeToSDLKey[DIK_NUMPAD2] = SDLK_KP2;
        aryScanCodeToSDLKey[DIK_NUMPAD3] = SDLK_KP3;
        aryScanCodeToSDLKey[DIK_NUMPAD0] = SDLK_KP0;
        aryScanCodeToSDLKey[DIK_DECIMAL] = SDLK_KP_PERIOD;
        aryScanCodeToSDLKey[DIK_F11] = SDLK_F11;
        aryScanCodeToSDLKey[DIK_F12] = SDLK_F12;
        
        aryScanCodeToSDLKey[DIK_F13] = SDLK_F13;
        aryScanCodeToSDLKey[DIK_F14] = SDLK_F14;
        aryScanCodeToSDLKey[DIK_F15] = SDLK_F15;
        
        aryScanCodeToSDLKey[DIK_NUMPADEQUALS] = SDLK_KP_EQUALS;
        aryScanCodeToSDLKey[DIK_NUMPADENTER] = SDLK_KP_ENTER;
        aryScanCodeToSDLKey[DIK_RCONTROL] = SDLK_RCTRL;
        aryScanCodeToSDLKey[DIK_DIVIDE] = SDLK_KP_DIVIDE;
        aryScanCodeToSDLKey[DIK_SYSRQ] = SDLK_PRINT;
        aryScanCodeToSDLKey[DIK_RMENU] = SDLK_RALT;
        aryScanCodeToSDLKey[DIK_PAUSE] = SDLK_PAUSE;
        aryScanCodeToSDLKey[DIK_HOME] = SDLK_HOME;
        aryScanCodeToSDLKey[DIK_UP] = SDLK_UP;
        aryScanCodeToSDLKey[DIK_PRIOR] = SDLK_PAGEUP;
        aryScanCodeToSDLKey[DIK_LEFT] = SDLK_LEFT;
        aryScanCodeToSDLKey[DIK_RIGHT] = SDLK_RIGHT;
        aryScanCodeToSDLKey[DIK_END] = SDLK_END;
        aryScanCodeToSDLKey[DIK_DOWN] = SDLK_DOWN;
        aryScanCodeToSDLKey[DIK_NEXT] = SDLK_PAGEDOWN;
        aryScanCodeToSDLKey[DIK_INSERT] = SDLK_INSERT;
        aryScanCodeToSDLKey[DIK_DELETE] = SDLK_DELETE;
        aryScanCodeToSDLKey[DIK_LWIN] = SDLK_LMETA;
        aryScanCodeToSDLKey[DIK_RWIN] = SDLK_RMETA;
        aryScanCodeToSDLKey[DIK_APPS] = SDLK_MENU;
    }
    
    LRESULT CALLBACK ExtendWndProc(HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
    {
        switch (msg) {
            case WM_SYSCOMMAND:
                switch (wparam) {
                    case WM_SC_SYSMENUABOUT: {
                            char ary_cBufferTmp[1024];
                            auto const fileAbout = fopen(szAboutFile, "r");
                            if (!fileAbout) return 0;
                            
                            auto const sRead = fread(ary_cBufferTmp, sizeof(*ary_cBufferTmp)
                                                    ,LENGTH_OF(ary_cBufferTmp) - 1, fileAbout);
                            fclose(fileAbout);
                            
                            ary_cBufferTmp[sRead] = '\0'; // Cap it
                            if (!*ary_cBufferTmp) return 0;
                            
                            ::MessageBox(hwnd, ary_cBufferTmp, "About", MB_OK);
                        }
                        return 0;
                        
                    case WM_SC_SYSMENUPASTEWIN32     : SysMenu::PasteClipboard       (true); return 0;
                    case WM_SC_SYSMENUREDIRECTPRINTER: SysMenu::PrinterRedirectToggle(true); return 0;
                    case WM_SC_SYSMENUPRINTSCREEN    : SysMenu::PrintScreen          (true); return 0;
                }
                break;
        }
        
        return ::CallWindowProc(fnSDLDefaultWndProc, hwnd, msg, wparam, lparam);
    }
#endif
    
    void GenKBStroke(const uint8_t scancode, const bool bDepressed, const SDLMod keymods)
    {
        auto const sdlkey = aryScanCodeToSDLKey[scancode];
        if (sdlkey == SDLK_UNKNOWN) return;
        
        SDL_Event evntKeyStroke = {0};
        evntKeyStroke.type                = bDepressed ? SDL_KEYDOWN : SDL_KEYUP;
        evntKeyStroke.key.keysym.scancode = scancode;
        evntKeyStroke.key.keysym.sym      = sdlkey;
        evntKeyStroke.key.keysym.mod      = keymods;
        evntKeyStroke.key.keysym.unicode  = 0;
        evntKeyStroke.key.state           = bDepressed ? SDL_PRESSED : SDL_RELEASED;
        
        MAPPER_CheckEvent(&evntKeyStroke);
    }
    
    void PrinterRedirectRefresh()
    {
        assert(printerRedirectPort);
        if (!printerRedirectPort) return;
#if defined(WIN32)
        MENUITEMINFO menuItemInfo = {0};
        menuItemInfo.cbSize = sizeof(menuItemInfo);
        menuItemInfo.fMask  = MIIM_STATE;
        menuItemInfo.fState = printerRedirect ? MFS_CHECKED : MFS_UNCHECKED;
        auto const ok = SetMenuItemInfo(::GetSystemMenu(WindowGet(), FALSE)
                                       ,WM_SC_SYSMENUREDIRECTPRINTER, FALSE, &menuItemInfo);
        assert(ok);
#else
        #error Impl for other platforms
#endif
    }
    
    void PrinterRedirectSet(bool const enabled)
    {
        printerRedirect = enabled;
        PrinterRedirectRefresh();
        
        for (uint32_t i = 0; i < parallelPortsMax; ++i) {
            if (!parallelPorts[i]) continue;
            parallelPorts[i]->RedirectEnable(enabled, printerRedirectPort);
        }
    } 
}


namespace SysMenu {

void Setup()
{
    assert(!setupDone);
    if (setupDone) return;
    setupDone = true;
    
#if defined(WIN32)
    PasteInitMapSCToSDLKey();
    
    auto const window  = WindowGet();
    auto const sysMenu = ::GetSystemMenu(window, FALSE);
    ::AppendMenu(sysMenu, MF_SEPARATOR, NULL, "");
    ::AppendMenu(sysMenu, MF_STRING, WM_SC_SYSMENUPASTEWIN32, "Paste From Windows (Ctrl+F9)");
    if (printerRedirectPort) ::AppendMenu(sysMenu, MF_STRING, WM_SC_SYSMENUREDIRECTPRINTER, "Print to dbDOS clipboard (Ctrl+F8)");
    if (printScreenPort    ) ::AppendMenu(sysMenu, MF_STRING, WM_SC_SYSMENUPRINTSCREEN    , "Print screen to dbDOS clipboard (Ctrl+PrnScr)");
    ::AppendMenu(sysMenu, MF_STRING, WM_SC_SYSMENUABOUT, "About...");
    
    PrinterRedirectRefresh();
    
    assert(!fnSDLDefaultWndProc);
    fnSDLDefaultWndProc = (WNDPROC)::GetWindowLongPtr(window, GWLP_WNDPROC);
    ::SetWindowLongPtr(window, GWLP_WNDPROC, (LONG_PTR)&ExtendWndProc);
#else
    #error Impl for other platforms
#endif
}

void PasteClipboard(bool const pressed)
{
    if (!pressed) return;
    
#if defined(WIN32)
    auto const window = WindowGet();
    if (!window                 ) return;
    if (!::OpenClipboard(window)) return;
    
    if (::IsClipboardFormatAvailable(CF_TEXT)) {
        if (auto const contents = ::GetClipboardData(CF_TEXT)) {
            if (auto const clipboard = (char const*)::GlobalLock(contents)) {
                // Create a copy of the string, and filter out Linefeed characters (ASCII '10')
                auto const clipboardLen = strlen(clipboard);
                for (size_t i = 0; i < clipboardLen; ++i) {
                    if (clipboard[i] == 0x0A) continue; //< Skip linefeeds
                    pasteBuffer.push(clipboard[i]);
                }
                
                ::GlobalUnlock(contents); 
            }
        }
    }
    
    ::CloseClipboard();
#else
    #error Impl for other platforms
#endif
}

void PasteClipboardTick()
{
    // Disabled multiple characters per dispatch b/c occasionally
    // keystrokes get lost in the spew. (Prob b/c of DI usage on Win32, sadly..)
    static Bitu pasteTicker = 0;
    if ((pasteTicker++ % 10) != 0) return;
    if (pasteBuffer.empty()      ) return;
    
    auto const character = pasteBuffer.front();
    
#if defined(WIN32)
    auto const virtualKey    = VkKeyScan(character); //< If it fails then MapVirtK will also fail, so no bail yet
    auto const scanCodeShift = MapVirtualKey(VK_SHIFT          , MAPVK_VK_TO_VSC);
    auto const scanCodeCtrl  = MapVirtualKey(VK_CONTROL        , MAPVK_VK_TO_VSC);
    auto const scanCodeAlt   = MapVirtualKey(VK_MENU           , MAPVK_VK_TO_VSC);
    auto const scanCode      = MapVirtualKey(LOBYTE(virtualKey), MAPVK_VK_TO_VSC);
    auto const modShift      = !!(virtualKey & 0x100);
    auto const modCtrl       = !!(virtualKey & 0x200);
    auto const modAlt        = !!(virtualKey & 0x400);
#else
    #error Impl for other platforms
#endif
    
    if (scanCode) {
        // Could be made more efficient, but would require tracking of more state,
        // so let's forgot that for now...
        auto const sdlmModsOn     = SDL_GetModState();
        auto const modOnShift     = !!(sdlmModsOn & (KMOD_LSHIFT | KMOD_RSHIFT));
        auto const modOnCtrl      = !!(sdlmModsOn & (KMOD_LCTRL  | KMOD_RCTRL ));
        auto const modOnAlt       = !!(sdlmModsOn & (KMOD_LALT   | KMOD_RALT  ));
        auto const modToggleShift = (modShift != modOnShift);
        auto const modToggleCtrl  = (modCtrl  != modOnCtrl );
        auto const modToggleAlt   = (modAlt   != modOnAlt  );
        auto const sdlmMods       = (SDLMod)((sdlmModsOn & ~(KMOD_LSHIFT | KMOD_RSHIFT |
                                                             KMOD_LCTRL  | KMOD_RCTRL  |
                                                             KMOD_LALT   | KMOD_RALT   )) |
                                             (modToggleShift ? KMOD_LSHIFT : 0) |
                                             (modToggleCtrl  ? KMOD_LCTRL  : 0) |
                                             (modToggleAlt   ? KMOD_LALT   : 0));
        
        /// \fixme Should check if key is already pressed or not so it can toggle press
        ///        but since we don't actually have any mappings from VK/SC to DI codes
        ///        (which SDL (can) use(s) internally as actually scancodes), we can't
        ///        actually check that ourselves, sadly...
        auto const strokesRequired = 2 + (modToggleShift ? 2 : 0)  //< key + shift up/down
                                       + (modToggleCtrl  ? 2 : 0)  //<     + ctrl  up/down
                                       + (modToggleAlt   ? 2 : 0); //<     + atl   up/down
        if (KEYBOARD_BufferSpaceAvail() < (strokesRequired + kPasteMinBufExtra))
            return;
        
        if (modToggleShift) GenKBStroke(scanCodeShift, !modOnShift, sdlmMods);
        if (modToggleCtrl ) GenKBStroke(scanCodeCtrl , !modOnCtrl , sdlmMods);
        if (modToggleAlt  ) GenKBStroke(scanCodeAlt  , !modOnAlt  , sdlmMods);
                            GenKBStroke(scanCode     ,  true      , sdlmMods);
                            GenKBStroke(scanCode     ,  false     , sdlmMods);
        if (modToggleShift) GenKBStroke(scanCodeShift,  modOnShift, sdlmMods);
        if (modToggleCtrl ) GenKBStroke(scanCodeCtrl ,  modOnCtrl , sdlmMods);
        if (modToggleAlt  ) GenKBStroke(scanCodeAlt  ,  modOnAlt  , sdlmMods);
    }
    
    pasteBuffer.pop();
}

void PrinterRedirectToggle(bool const pressed)
{
    if (!pressed) return;
    PrinterRedirectSet(!printerRedirect);
}



void PrinterRedirectPortSet(uint32_t const port)
{
    // Only set once since we modify the system menu, and we might have redirected prints in progress.
    assert(port != 0 );
    assert(!setupDone);
    if (setupDone          ) return;
    if (printerRedirectPort) return;
    
    printerRedirectPort = port;
    //LOG_MSG("Print to dbDOS clipboard: Received redirect port %u", port);
}

void PrintScreenPortSet(uint32_t const port)
{
    // Only set once since we modify the system menu, and we might have redirected prints in progress.
    assert(port != 0 );
    assert(!setupDone);
    if (setupDone      ) return;
    if (printScreenPort) return;
    
    printScreenPort = port;
    //LOG_MSG("Print Screen: Received redirect port %u", port);
}

void PrintScreen(bool const pressed)
{
    if (!pressed) return;
    CaptureState |= CAPTURE_SCREEN;
}

void PrintScreenSender::png_rw(png_structp const png, png_bytep const data, png_size_t const size)
{
    auto const self = reinterpret_cast<PrintScreenSender*>(png_get_io_ptr(png));
    assert(self);
    self->_socket.WriteRaw(data, size);
}

void PrintScreenSender::png_flush(png_structp const png)
{
    auto const self = reinterpret_cast<PrintScreenSender*>(png_get_io_ptr(png));
    assert(self);
    (void)self; //< Do nothing. We flush at death.
}

PrintScreenSender::PrintScreenSender(png_structp png_ptr) : _socket("127.0.0.1", printScreenPort)
                                                          , _structsOk(false)
{
    if (!png_ptr          ) return;
    if (!_socket.IsValid()) return;
    
    png_set_write_fn(png_ptr, this, png_rw, png_flush);
    _structsOk = true;
}


#ifndef BI_RGB
    #define BI_RGB 0
#endif
void PrintScreenSend(uint32_t const width, uint32_t const height, uint32_t const bpp, uint32_t const pitch
                    ,uint8_t  const data[/*pitch * height*/], uint8_t const pallete[/*256 * 4*/])
{
    if (!printScreenPort) return;
    
    TCPSocket socket("127.0.0.1", printScreenPort);
    if (!socket.IsValid()) {
        LOG_MSG("Print Screen Error: Failed to connect to dbDOS launcher. %s"
               ,socket.ErrorGet().c_str());
        return;
    }
    
    assert(pitch == ((((bpp * width) + 31) / 32) * 4));
    auto const palletised = (bpp <= 8);
    auto const imageSize  = pitch * height;
    // Send it over the wire as a packed DIB
    struct {
        DWORD      biSize;
        LONG       biWidth;
        LONG       biHeight;
        WORD       biPlanes;
        WORD       biBitCount;
        DWORD      biCompression;
        DWORD      biSizeImage;
        LONG       biXPelsPerMeter;
        LONG       biYPelsPerMeter;
        DWORD      biClrUsed;
        DWORD      biClrImportant;
    } const header = {sizeof(header), width, height, 1, bpp, BI_RGB, imageSize, 0, 0
                     ,palletised ? 256 : 0, 0};
    socket.WriteRaw(header); //< \note Not safe if dest != same endianness
    
    if (palletised)
        socket.WriteRaw(pallete, 256 * 4); //< \note Safe. Pallete datum is oclet-sized.
    
    for (auto i = height; i > 0; --i) //< bottom -> top
        socket.WriteRaw(data + pitch * (i - 1), pitch); //< \note Not safe if dest != same endianness
}

}

